/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994-1996,1999  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: curses.c,v 1.33 1999/05/16 06:56:25 mhw Exp $
 */

#include "netris.h"
#include <sys/types.h>
#include <unistd.h>
#include <curses.h>
#include <string.h>
#include <stdlib.h>

#ifdef NCURSES_VERSION
# define HAVE_NCURSES
#endif

#ifdef HAVE_NCURSES
static struct
{
	BlockType type;
	short color;
} myColorTable[] =
{
	{ BT_white,		COLOR_WHITE },
	{ BT_blue,		COLOR_BLUE },
	{ BT_magenta,	COLOR_MAGENTA },
	{ BT_cyan,		COLOR_CYAN },
	{ BT_yellow,	COLOR_YELLOW },
	{ BT_green,		COLOR_GREEN },
	{ BT_red,		COLOR_RED },
	{ BT_none, 0 }
};
#endif

ExtFunc void PlotBlock1(int scr, int y, int x, BlockType type);
static MyEventType KeyGenFunc(EventGenRec *gen, MyEvent *event);

static EventGenRec keyGen =
		{ NULL, 0, FT_read, STDIN_FILENO, KeyGenFunc, EM_key };

static int boardYPos[MAX_SCREENS], boardXPos[MAX_SCREENS];
static int statusYPos, statusXPos;
static int haveColor;

static char *term_vi;	/* String to make cursor invisible */
static char *term_ve;	/* String to make cursor visible */

ExtFunc void InitScreens(void)
{
	MySigSet oldMask;

	GetTermcapInfo();

	/*
	 * Block signals while initializing curses.  Otherwise a badly timed
	 * Ctrl-C during initialization might leave the terminal in a bad state.
	 */
	BlockSignals(&oldMask, SIGINT, 0);
	initscr();

#ifdef CURSES_HACK
	{
		extern char *CS;

		CS = 0;
	}
#endif

#ifdef HAVE_NCURSES
	haveColor = Game.color && has_colors();
	if (haveColor) {
		int i = 0;

		start_color();
		for (i = 0; myColorTable[i].type != BT_none; ++i)
			init_pair(myColorTable[i].type, COLOR_BLACK,
					myColorTable[i].color);
	}
#else
	haveColor = 0;
#endif

	AtExit(CleanupScreens);
	RestoreSignals(NULL, &oldMask);

	cbreak();
	noecho();
	OutputTermStr(term_vi, 0);
	AddEventGen(&keyGen);

	move(0, 1);
	addstr("NETRIS ");
	addstr(version_string);
	{
		int rows, cols;

		getmaxyx(stdscr, rows, cols);
		move(0, cols - 48);
		addstr("(C)1994-1996,1999 Mark H. Weaver, (C)2002 Shiar");
//		addstr("    \"netris -h\" for more info");
	}
	statusYPos = 22;
	statusXPos = 0;
}

ExtFunc void CleanupScreens(void)
{
	RemoveEventGen(&keyGen);
	endwin();
	OutputTermStr(term_ve, 1);
}

ExtFunc void GetTermcapInfo(void)
{
	char *term, *buf, *data;
	int bufSize = 10240;

	if (!(term = getenv("TERM")))
		return;
	if (tgetent(scratch, term) == 1) {
		/*
		 * Make the buffer HUGE, since tgetstr is unsafe.
		 * Allocate it on the heap too.
		 */
		data = buf = malloc(bufSize);

		/*
		 * There is no standard include file for tgetstr, no prototype
		 * definitions.  I like casting better than using my own prototypes
		 * because if I guess the prototype, I might be wrong, especially
		 * with regards to "const".
		 */
		term_vi = (char *)tgetstr("vi", &data);
		term_ve = (char *)tgetstr("ve", &data);

		/* Okay, so I'm paranoid; I just don't like unsafe routines */
		if (data > buf + bufSize)
			fatal("tgetstr overflow, you must have a very sick termcap");

		/* Trim off the unused portion of buffer */
		buf = realloc(buf, data - buf);
	}

	/*
	 * If that fails, use hardcoded vt220 codes.
	 * They don't seem to do anything bad on vt100's, so
	 * we'll try them just in case they work.
	 */
	if (!term_vi || !term_ve) {
		static char *vts[] = {
				"vt100", "vt101", "vt102",
				"vt200", "vt220", "vt300",
				"vt320", "vt400", "vt420",
				"screen", "xterm", NULL };
		int i;

		for (i = 0; vts[i]; i++)
			if (!strcmp(term, vts[i]))
			{
				term_vi = "\033[?25l";
				term_ve = "\033[?25h";
				break;
			}
	}
	if (!term_vi || !term_ve)
		term_vi = term_ve = NULL;
}

ExtFunc void OutputTermStr(char *str, int flush)
{
	if (str) {
		fputs(str, stdout);
		if (flush)
			fflush(stdout);
	}
}

ExtFunc void DrawField(int scr)
{
	{
		int y, x;

		for (y = Players[scr].boardVisible - 1; y >= 0; --y) {
			mvaddch(boardYPos[scr] - y, boardXPos[scr] - 1,
				Game.ascii ? '|' : ACS_VLINE); //left
			mvaddch(boardYPos[scr] - y,
				boardXPos[scr] + 2 * Players[scr].boardWidth,
				Game.ascii ? '|' : ACS_VLINE); //right
		}
		move(2, boardXPos[scr] - 1); //top
		addch(Game.ascii ? '+' : ACS_ULCORNER);
		for (x = Players[scr].boardWidth * 2 - 1; x >= 0; --x)
			addch(Game.ascii ? '-' : ACS_HLINE);
		addch(Game.ascii ? '+' : ACS_URCORNER);
		move(boardYPos[scr] + 1, boardXPos[scr] - 1); //bottom
		addch(Game.ascii ? '+' : ACS_LLCORNER);
		for (x = Players[scr].boardWidth * 2 - 1; x >= 0; --x)
			addch(Game.ascii ? '-' : ACS_HLINE);
		addch(Game.ascii ? '+' : ACS_LRCORNER);
	} //draw field grid

	{
		char userstr[300];

		sprintf(userstr, "%s <%s>", Players[scr].name, Players[scr].host);
		userstr[20 - 7*((Players[scr].flags & SCF_usingRobot) != 0)
			- 5*((Players[scr].flags & SCF_fairRobot) != 0)] = 0;
		mvaddstr(1, boardXPos[scr], userstr);

		if (Players[scr].flags & SCF_usingRobot) {
			addstr((Players[scr].flags & SCF_fairRobot)
				? "(fair robot)" : "(robot)");
		}
	} //display playername/host
} //DrawScreen

ExtFunc void InitFields()
{
	int scr, prevscr;

	boardXPos[me] = 1;
	boardYPos[me] = 22;
	prevscr = me;
	for (scr = 1; scr <= totalPlayers + 1; scr++) if (scr != me) {
		boardXPos[scr] =
			boardXPos[prevscr] + 2 * Players[prevscr].boardWidth + 3;
		if (prevscr == me)
			boardXPos[scr] += 24; //scorebar
		boardYPos[scr] = 22;
		prevscr = scr;
	}
	statusXPos = 2 * Players[me].boardWidth + 3;
} //InitScreen

ExtFunc void CleanupScreen(int scr)
{
}

ExtFunc void PlotBlock1(int scr, int y, int x, BlockType type)
{
	int colorIndex = abs(type);

	move(y, x);

	if (type == BT_none)
		addstr("  ");
	else
	{
		if (Game.standout)
		{
#ifdef HAVE_NCURSES
			if (haveColor)
				attrset(COLOR_PAIR(colorIndex));
			else
#endif
				standout();
		}

		addstr(type ? "[]" : "$$");
		standend();
	}
}

ExtFunc void PlotBlock(int scr, int y, int x, BlockType type)
{
	if (y >= 0 && y < Players[scr].boardVisible &&
		x >= 0 && x < Players[scr].boardWidth)
	  PlotBlock1(scr, boardYPos[scr] - y, boardXPos[scr] + 2 * x, type);
}

ExtFunc void PlotUnderline(int scr, int x, int flag)
{
  move(boardYPos[scr] + 1, boardXPos[scr] + 2 * x);
  if (Game.ascii)
	addstr(flag ? "==" : "--");
  else {
	addch(flag ? ACS_BTEE : ACS_HLINE);
	addch(flag ? ACS_BTEE : ACS_HLINE);
  }
}

ExtFunc void ShowDisplayInfo(void)
{
  move(statusYPos - 9, statusXPos);
  printw("Seed: %010d", Game.seed);
  //	move(statusYPos - 8, statusXPos);
  //	printw("Speed: %dms ", speed / 1000);
  if (robotEnable) {
	move(statusYPos - 6, statusXPos);
	if (fairRobot)
	  addstr("Controlled by a fair robot");
	else
	  addstr("Controlled by a robot");
	//		clrtoeol();
  }
  if (Players[1].flags & SCF_usingRobot) {
	move(statusYPos - 5, statusXPos);
	if (Players[1].flags & SCF_fairRobot)
	  addstr("The opponent is a fair robot");
	else
	  addstr("The opponent is a robot");
	//		clrtoeol();
  }
}

ExtFunc void ShowScore(int scr, struct _Score score)
{
	float timer;

	move(6, statusXPos); addstr("Next:          ");
	move(7, statusXPos + 7);    addstr("        ");
	ShapeIterate(Players[scr].nextShape, scr,
	  ShapeToNetNum(Players[scr].nextShape) == 15 ? 13 : 14,
	  statusXPos / 2 + 5, 1, GlanceFunc, NULL);
	move(statusYPos - 21 + 1, statusXPos);
	printw("Score:%6d  level: %2d", score.score, score.level);
	move(statusYPos - 20 + 1, statusXPos);
	timer = CurTimeval() / 1e6;
	printw("Lines:%6d", score.lines);
	if (timer > 4) {
		printw("  ppm:%5.1f", score.drops * 60 / timer);
		move(statusYPos - 18, statusXPos);
		if (score.lines > 0)
			printw("yield:  %3d%%", 100 * score.adds / score.lines);
		else addstr("            ");
		printw("  apm:%5.1f", score.adds * 60 / timer);
	}
}

ExtFunc void ShowPause(int pausedByMe, int pausedByThem)
{
	move(statusYPos - 3, statusXPos);
	if (pausedByThem)
		addstr("Game paused by opponent");
	else
		addstr("                       ");
	move(statusYPos - 2, statusXPos);
	if (pausedByMe)
		addstr("Game paused by you");
	else
		addstr("                  ");
}

ExtFunc void Message(char *s)
{
	static int line = 0;

//	move(statusYPos - 20 + line, statusXPos);
	move(statusYPos + 2 + line, 1);
	addstr(s);	/* XXX Should truncate long lines */
	clrtoeol();
	line = (line + 1) % 10;
//	move(statusYPos - 20 + line, statusXPos);
	move(statusYPos + 2 + line, 1);
	clrtoeol();
}

ExtFunc void ShowTime(void)
{
	move(statusYPos, statusXPos);
	printw("Timer: %.0f ", CurTimeval() / 1e6);
	move(boardYPos[0] + 1, boardXPos[0] + 2 * Players[0].boardWidth + 1);
//	refresh();
}

ExtFunc void ScheduleFullRedraw(void)
{
	touchwin(stdscr);
}

static MyEventType KeyGenFunc(EventGenRec *gen, MyEvent *event)
{
	if (MyRead(gen->fd, &event->u.key, 1))
		return E_key;
	else
		return E_none;
}

/*
 * vi: ts=4 ai
 * vim: noai si
 */
