/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994,1995,1996  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: util.c,v 1.29 1999/05/16 06:56:33 mhw Exp $
 */

#include "netris.h"
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netdb.h>
#include <errno.h>
#include <setjmp.h>

static MyEventType AlarmGenFunc(EventGenRec *gen, MyEvent *event);

static EventGenRec alarmGen =
		{ &alarmGen, 0, FT_read, -1, AlarmGenFunc, EM_alarm };
static EventGenRec *nextGen = &alarmGen;

static sigjmp_buf close_env;

static int myRandSeed = 1;

static long baseTimeval;

ExtFunc void AtExit(void (*handler)(void))
{
#ifdef HAS_ON_EXIT
	on_exit((void *)handler, NULL);
#else
	atexit(handler);
#endif
}

///////////// HELP MESSAGES /////////////

ExtFunc void Header(void)
{
	fprintf(stderr,
	  "NETRIS %s\t(c) 1994-1996,1999 Mark H. Weaver <mhw@netris.org>\n"
	  "          \t(c) 2002 Shiar <shiar@shiar.org>\n\n",
	  version_string);
} //Header

ExtFunc void Usage(void)
{
	Header();
	fprintf(stderr,
	  "Usage: netris <options>\n"
	  "\n"
	  "  -h, --help\t\tPrint this usage information\n"
	  "  -H, --info\t\tShow distribution and warranty information\n"
	  "  -R, --rules\t\tShow game rules\n"
	  "\n"
	  "  -S, --slowterm\tDisable inverse/bold/color for slow terminals\n"
	  "  -a, --ascii\t\tUse ascii characters\n"
	  "  -C, --color=0\t\tDisable color\n"
	  "\n"
	  "  -w, --wait\t\tWait for connection\n"
	  "  -c, --connect <host>\tInitiate connection\n"
	  "  -p, --port <port>\tSet port number (default is %d)\n"
	  "\n"
	  "  -s, --seed <seed>\tStart with given random seed\n"
	  "  -i, --speed <sec>\tSet the initial step-down interval, in seconds\n"
	  "  -l, --level <lvl>\tBegin at a higher level (can be used as handicap)\n"
	  "  -k, --keys <keys>\tRemap keys (default is \"%s\" for cursors)\n"
	  "  -d, --dropmode\tDrops go into drop mode\n"
	  "  -D, --instadrop\tInstant drop\n"
	  "\n"
	  "  -r, --robot <cmd>\tExecute program to control the game instead of keyboard\n"
	  "  -F, --fair-robot\tUse fair robot interface\n"
	  "\n", DEFAULT_PORT, DEFAULT_KEYS);
}

ExtFunc void DistInfo(void)
{
	Header();
	fprintf(stderr,
	  "This program is free software; you can redistribute it and/or modify\n"
	  "it under the terms of the GNU General Public License as published by\n"
	  "the Free Software Foundation; either version 2 of the License, or\n"
	  "(at your option) any later version.\n"
	  "\n"
	  "This program is distributed in the hope that it will be useful,\n"
	  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	  "GNU General Public License for more details.\n"
	  "\n"
	  "You should have received a copy of the GNU General Public License\n"
	  "along with this program; if not, write to the Free Software\n"
	  "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
	  "\n");
} //DistInfo

ExtFunc void Rules(void)
{
	Header();
	fprintf(stderr,
	  "One player mode\n"
	  "---------------\n"
	  "Good old Tetris. Scoring is like on the GameBoy version (so try to\n"
	  "remove as many lines at once as you can). After removing ten lines\n"
	  "you go to the next level, which will be faster thus making the game\n"
	  "harder to play.\n"
	  "\n"
	  "Two player mode\n"
	  "---------------\n"
	  "It's just like normal T*tris except that when you clear more than\n"
	  "one row with a single piece, the other player receives penalty lines\n"
	  "For clearing 2, 3 or 4 rows, respectively 1, 2 or 4 junk rows will\n"
	  "be added to the bottom of your opponent's board respectively.\n"
	  "The junk rows have exactly one empty column, which will line up for\n"
	  "multiple rows.\n"
	  "\n"
	  "The longest surviving player wins the game.\n"
	  "\n");
} //Rules

///////////// RANDOM /////////////

ExtFunc void InitUtil(void)
{
	if (Game.seed)
		SRandom(Game.seed);
	else
		SRandom(time(0));
	if (sigsetjmp(close_env, 1)) exit(0);
	signal(SIGINT, CatchInt);
	ResetBaseTime();
} //InitUtil

/*
 * My really crappy random number generator follows
 * Should be more than sufficient for our purposes though
 */
ExtFunc void SRandom(int seed)
{
	Game.seed = seed;
	myRandSeed = seed % 31751 + 1;
} //SRandom

ExtFunc int Random(int min, int max1)
{
	myRandSeed = (myRandSeed * 31751 + 15437) % 32767;
	return myRandSeed % (max1 - min) + min;
} //Random

///////////// I/O /////////////

ExtFunc int MyRead(int fd, void *data, int len)
{
	int result, left;

	left = len;
	while (left > 0) {
		result = read(fd, data, left);
		if (result > 0) {
			data = ((char *)data) + result;
			left -= result;
		}
		else if (errno != EINTR)
			return result;
	}
	return len;
} //MyRead

ExtFunc int MyWrite(int fd, void *data, int len)
{
	int result, left;

	left = len;
	while (left > 0) {
		result = write(fd, data, left);
		if (result > 0) {
			data = ((char *)data) + result;
			left -= result;
		}
		else if (errno != EINTR)
			return result;
	}
	return len;
} //MyWrite

///////////// TIME /////////////

ExtFunc void NormalizeTime(struct timeval *tv)
{
	tv->tv_sec += tv->tv_usec / 1000000;
	tv->tv_usec %= 1000000;
	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		--tv->tv_sec;
	}
}

ExtFunc void CatchInt(int sig)
{
	siglongjmp(close_env, 1);
}

ExtFunc void CatchAlarm(int sig)
{
	alarmGen.ready = 1;
	signal(SIGALRM, CatchAlarm);
}

static MyEventType AlarmGenFunc(EventGenRec *gen, MyEvent *event)
{
	return E_alarm;
}

ExtFunc void SetTimeval(struct timeval *tv, long usec)
{
	tv->tv_sec = usec / 1000000;
	tv->tv_usec = usec % 1000000;
} //SetTimeval

ExtFunc long GetTimeval(struct timeval *tv)
{
	return tv->tv_sec * 1000000 + tv->tv_usec;
} //GetTimeval

ExtFunc long AbsTimeval(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return GetTimeval(&tv);
} //CurTimeval

ExtFunc void ResetBaseTime(void)
{
	baseTimeval = AbsTimeval();
} //ResetBaseTime

ExtFunc void PauseTime(void)
{
	baseTimeval -= AbsTimeval();
} //PauseTime

ExtFunc void ResumeTime(void)
{
	baseTimeval += AbsTimeval();
} //ResumeTime

ExtFunc long CurTimeval(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return GetTimeval(&tv) - baseTimeval;
} //CurTimeval

static long SetITimer1(long interval, long value)
{
	struct itimerval it, old;

	SetTimeval(&it.it_interval, interval);
	SetTimeval(&it.it_value, value);
	if (setitimer(ITIMER_REAL, &it, &old) < 0)
		die("setitimer");
	signal(SIGALRM, CatchAlarm);
	return GetTimeval(&old.it_value);
}

ExtFunc long SetITimer(long interval, long value)
{
	long old;

	old = SetITimer1(0, 0);
	alarmGen.ready = 0;
	SetITimer1(interval, value);
	return old;
}

///////////// ... /////////////

ExtFunc volatile void die(char *msg)
{
	perror(msg);
	exit(1);
}

ExtFunc volatile void fatal(char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

ExtFunc void BlockSignals(MySigSet *saved, ...)
{
	MySigSet set;
	va_list args;
	int sig;

	va_start(args, saved);
#ifdef HAS_SIGPROCMASK
	sigemptyset(&set);
#else
	set = 0;
#endif
	while ((sig = va_arg(args, int))) {
#ifdef HAS_SIGPROCMASK
		sigaddset(&set, sig);
#else
		sig |= sigmask(sig);
#endif
	}
#ifdef HAS_SIGPROCMASK
	sigprocmask(SIG_BLOCK, &set, saved);
#else
	*saved = sigblock(set);
#endif
	va_end(args);
}

ExtFunc void RestoreSignals(MySigSet *saved, MySigSet *set)
{
#ifdef HAS_SIGPROCMASK
	sigprocmask(SIG_SETMASK, set, saved);
#else
	if (saved)
		*saved = sigsetmask(*set);
	else
		sigsetmask(*set);
#endif
}

///////////// EVENTS /////////////

ExtFunc void AddEventGen(EventGenRec *gen)
{
	assert(gen->next == NULL);
	gen->next = nextGen->next;
	nextGen->next = gen;
} //AddEventGen

ExtFunc void RemoveEventGen(EventGenRec *gen)
{
	// assert(gen->next != NULL);	/* Be more forgiving, for SIGINTs */
	if (gen->next) {
		while (nextGen->next != gen)
			nextGen = nextGen->next;
		nextGen->next = gen->next;
		gen->next = NULL;
	}
} //RemoveEventGen

ExtFunc MyEventType WaitMyEvent(MyEvent *event, int mask)
{ //poll
	int i, retry = 0;
	fd_set fds[FT_len];
	EventGenRec *gen;
	int result, anyReady, anySet;
	struct timeval tv;

	for (;;) {
		for (i = 0; i < FT_len; ++i)
			FD_ZERO(&fds[i]);
		anyReady = anySet = 0;
		gen = nextGen;
		do {
			if (gen->mask & mask) {
				if (gen->ready)
					anyReady = 1;
				if (gen->fd >= 0) {
					FD_SET(gen->fd, &fds[gen->fdType]);
					anySet = 1;
				}
			}
			gen = gen->next;
		} while (gen != nextGen);
		if (anySet) {
			tv.tv_sec = 0;
			tv.tv_usec = (retry && !anyReady) ? 500000 : 0;
			result = select(FD_SETSIZE, &fds[FT_read], &fds[FT_write],
					&fds[FT_except], anyReady ? &tv : NULL);
		}
		else {
			if (retry && !anyReady)
				sleep(1);
			result = 0;
		}
		gen = nextGen;
		do {
			if ((gen->mask & mask)
					&& (gen->ready || (result > 0 && gen->fd >= 0
						&& FD_ISSET(gen->fd, &fds[gen->fdType])))) {
				gen->ready = 0;
				event->type = gen->func(gen, event);
				if (event->type != E_none) {
					nextGen = gen->next;
					return event->type;
				}
			}
			gen = gen->next;
		} while (gen != nextGen);
		retry = 1;
	}
} //WaitMyEvent

/*
 * vi: ts=4 ai
 * vim: noai si
 */
